
;	V2080.ASM
;	FIRST ISSUE: DECEMBER 4, 1985
;	8080 EMULATOR FOR V20
;	WRITTEN BY MICHAEL DAY
;
;History:
;
;	1st Update as of 5 Oct 1986
;	By Michael Day
;	Converted to generic operation to allow use with device drivers
;	Fixed bug about not running properly on Compaq computer
;	(flipped speed to slow when program was run
;	 caused by interrupt vector conflicts.)
;
;	2nd update as of 15 Nov 1987
;	Fixed bug per Ward Christensen about not clearing 
;	8080 flag from stack
;
;	3rd update as of 8 Jan 1988
;	Added detection for V20/V30 chip.
;	Operation is disallowed if chip is not present.
;	An IF was used to control the code so that it can
;	be trapped out if it is a problem. 
;
;This program expects the Seattle Computer Products 8086
;assembler to be used to assemble it. Alternately, the
;Hawthorne Technology 8086 assembler can be used which is
;compatible to SASM.
;Or you can modify the program to work with MASM.
;The original cpemulator program was written in SASM which
;is why it is still in that assmebly style, besides which
;SASM is about ten times faster than MASM on small files 
;like this.


	ORG 100H

NO	EQU 0000H
YES	EQU 0FFFFH

CR	EQU 0DH
LF	EQU 0AH
BELL	EQU 07H	

EMADM3	 EQU NO 	;YES=EMULATE ADM3A, NO=NO EMULATION
NOEMADM3 EQU YES 	;OPPOSITE OF EMADM3 (BE SURE TO SET BOTH)

NECCHK	EQU YES		;TRUE=CHECK FOR EXISTANCE OF V20
			;MAY NOT WORK WITH 386, BUT WE'LL SEE

BRKEM	EQU 0FF0FH	;V20 CODE TO ENTER 8080 MODE
RETEM	EQU 0FDEDH	;V20 CODE TO RETURN TO NATIVE MODE
CALLN	EQU 0EDEDH	;V20 CODE TO CALL NATIVE FROM 8080 MODE

EMINT	EQU 68H		;THE INTERRUPT VECTOR WE USE TO ENTER 8080 MODE

RAMSEGMENT	EQU 300H	;3000H  12K FOR EMULATOR


;
; SET UP ALL VARIABLES FOR EMULATOR
; DO ALL ONE TIME INITILIZATION
;
;	INITALIZATION SECTION
;
	JMP INIT

ATRIB:	DB 0FH

CPMMSG:
     IF NOEMADM3
	DB 'This version of V2080 does not have terminal emulation. For terminal emulation',CR,LF
	DB 'use a screen device driver such as ANSI.SYS. See documentation for more info.',CR,LF
     ENDIF
     IF EMADM3
	DB 'This version of V2080 supports ADM3A terminal emulation. It will bypass any',CR,LF
	DB 'device driver that you might have installed. See documentation for more info.',CR,LF
     ENDIF
CPMMSG2:
	DB CR,LF
	DB '     64K  V2080 CPeMulator  Version 3.21 as of  8 January, 1988',CR,LF
	DB '          Copyright (c) 1985, 1986, 1987, 1988  Michael Day',CR,LF
	DB CR,LF
	DB 'OPERATION: To run a program, type "V2080 FILENAME" at the MS-DOS prompt (CP/M',CR,LF
 	DB 'programs are assumed to have a filename extent of .CPM).  Use of this program',CR,LF
	DB 'requires that a NEC V20 chip be installed in the computer in place of the 8088',CR,LF
	DB 'mpu (or V30 for 80186 machines).  This will allow you to run 8080 based CP/M-80',CR,LF
	DB 'programs on your MS-DOS machine (within hardware and software limitations).',CR,LF
	DB 'DISCLAIMER: I make no warranties or claims of the fitness of this product for',CR,LF
	DB 'any particular purpose.  (As always, make backups first!)',CR,LF
	DB 'NOTICE: Users of this program are granted a license to make copies of this',CR,LF
	DB 'program for use on a private non-commercial basis.  A simple and painless',CR,LF
	DB 'commercial license is also available.  See documentation for more information.',CR,LF 
	DB 'No enforcement will be made of this policy, it is simply a (reasonable I hope)',CR,LF
	DB 'request.  Thank you for your cooperation!',CR,LF
	DB 0FFH

INTELMSG:
	DB 'A V20/V30 was not detected.  V2080 CPeMulator requires a NEC V20 or V30 mpu',CR,LF
	DB 'to operate. See below on how to obtain a 8086/286/386 version of CPeMulator.' ,CR,LF
	DB 0FFH

; This is a simple clear screen nothing fancy, just roll the screen up
BLNKSCR:
	DB CR,LF	;1
	DB CR,LF
	DB CR,LF	;3
	DB CR,LF
	DB CR,LF	;5
	DB CR,LF
	DB CR,LF	;7
	DB CR,LF
	DB CR,LF	;9
	DB CR,LF
	DB CR,LF	;11
	DB CR,LF
	DB CR,LF	;13
	DB CR,LF
	DB CR,LF	;15
	DB CR,LF
	DB CR,LF	;17
	DB CR,LF
	DB CR,LF	;19
	DB CR,LF
	DB CR,LF	;21
	DB CR,LF
	DB CR,LF	;23
	DB CR,LF
	DB 0FFH

;-------------------------------------------------------------------
; BEG OF PROGRAM


	DB 'So, your looking around inside here are you? Tsk tsk tsk...',CR,LF
	DB 'If you are really interested, the source code *is* available',CR,LF
	DB 'Save yourself a lot of trouble and get real commented source',CR,LF
	DB 'code instead of trying to figure it all out with a debugger?',CR,LF
	DB 0FFH

USERBOOT:
	INT 20H		;RETURN TO MSDOS


;GETSCREEN:
;	MOV AH,06	;OUTPUT THE CHAR VIA FUNCTION 6
;	INT 21H
;	RET

;OUTPUT CONTROL ROUTINES
;OUTPUT A CHARACTER TO THE SCREEN
;A SCREEN DEVICE DRIVER SHOULD BE USED TO EMULATE
;A TERMINAL IF DESIRED. SUCH AS USING ANSI.SYS 
;WHICH EMULATES A VT100 TYPE TERMINAL. 
;ALTERNATELY YOU CAN ENABLE EMADM3 TO TURN ON INTERNAL ADM3A EMULATION

     IF EMADM3
CURCONT:
	CMP DL,00H	;IGNORE NULLS
	JZ CURBYPASS

	MOV AL,[EQFLAG]		;WE GOT AN ESC IS IT CURSOR CONT?
	CMP AL,0FFH
	JZ GETROW
	CMP DL,'='
	JNZ CLRCURFLG
	MOV B,[EQFLAG],0FFH
CURBYPASS:
	RET

GETROW:
	MOV AL,[ROWFLAG]
	CMP AL,0FFH
	JZ GETCOL
	SUB DL,32
	INC DL
	MOV [ROWCNT],DL
	MOV B,[ROWFLAG],0FFH
	RET

GETCOL:
	SUB DL,32
	MOV DH,[ROWCNT]
	MOV AH,2
	MOV AL,0
	MOV BX,0
	INT 10H

CLRCURFLG:
	MOV B,[CURFLAG],00H
	MOV B,[EQFLAG],00H
	MOV B,[ROWFLAG],00H
	RET
     ENDIF

;
GETSCREEN:
     IF EMADM3
	AND DL,7FH		;ESC SEQUENCE?
	CMP B,[CURFLAG],0FFH
	JZ CURCONT

	CMP DL,00H		;NULL?
	JZ GETSCEXIT
	CMP DL,1BH		;ESC CHAR
	JZ SETESC
	CMP DL,1AH		;CLEAR SCREEN
	JZ CLEARSCX
	CMP DL,1EH		;HOME
	JZ SETHOMEX
	CMP DL,0BH		;CURSOR UP?
	JZ CURSORUPX
	CMP DL,17H		;CLEAR END OF SCREEN
	JZ CLREOSX
	CMP DL,18H		;CLEAR EOL
	JZ CLREOLX

	CMP DL,0CH		;CURSOR RIGHT?
	JNZ CHAROUTOK
	JMP CURRIGHT

GETSCEXIT:
	RET

CLEARSCX:
	JMP CLEARSC

SETHOMEX:
	JMP SETHOME

CURSORUPX:
	JMP CURSORUP

CLREOSX:
	JMP CLREOS

CLREOLX:
	JMP CLREOL


SETESC:
	MOV B,[CURFLAG],0FFH
	RET
     ENDIF

CHAROUTOK:
     IF EMADM3
	PUSH DX
	MOV AH,3	;FIND OUT WHERE WE ARE AT
	MOV AL,0
	MOV BX,0
	INT 10H
	MOV [LASTXY],DX
	POP DX
     ENDIF

	MOV AH,06		;NORMAL CHAR
	INT 21H

     IF EMADM3
	MOV AH,3
	MOV AL,0
	MOV BX,0
	INT 10H

	MOV BX,[LASTXY]
	CMP DH,0
	JZ FIXROW 
	CMP DH,24
	JC NOERASETOP
	CMP BH,24
	JC NOERASETOP
	CMP DL,0
	JZ ERASETOP
	CMP DL,BL
	JZ ERASETOP
NOERASETOP:
	RET
ERASETOP:
	MOV AH,7	;ERASE THE TOP LINE ON THE SCREEN
	MOV AL,0
	MOV CX,0
	MOV DH,0
	MOV DL,79
	SEG CS
	MOV BH,[ATRIB]
	INT 10H
	RET

FIXROW:
	MOV DH,1	;OFFSET BACK TO LINE 1 IF WAS AT 0
	MOV AH,2
	MOV AL,0
	MOV BX,0
	INT 10H
	RET
     ENDIF

	RET

     IF EMADM3
CLREOL:	MOV AH,3
	MOV AL,0
	MOV BX,0
	INT 10H
	MOV AL,0
	MOV AH,6
	SEG CS
	MOV BH,[ATRIB]
	MOV CX,DX
	MOV DL,79
	INT 10H
	RET

CLREOS:	CALL CLREOL
	MOV AH,3
	MOV AL,0
	MOV BX,0
	INT 10H
	MOV AH,6
	SEG CS
	MOV BH,[ATRIB]
	MOV AL,0
	MOV CX,DX
	INC CH
	MOV CL,0
	MOV DH,24
	MOV DL,79
	INT 10H
	RET

CLEARSC:
	MOV AH,6
	MOV AL,0
	MOV CX,0
	MOV DH,24
	MOV DL,79
	SEG CS
	MOV BH,[ATRIB]
	INT 10H		;FALL THRU TO SETHOME

SETHOME:
	MOV AH,2
	MOV AL,0
	MOV DH,0
	MOV DL,0
	INC DH		;LINE 1 IS TOP OF PAGE IF BLANK TOP 	
	MOV BX,0
	INT 10H
	RET

CURSORUP:
	MOV AH,3
	MOV AL,0
	MOV BX,0
	INT 10H
	DEC DH
	CMP DH,0
	JNZ ROWUPOK
	INC DH

ROWUPOK:
	MOV AH,2
	MOV AL,0
	MOV BX,0
	INT 10H
	RET


CURRIGHT:
	MOV AL,0
	MOV AH,3
	MOV BX,0
	INT 10H
	INC DL
	CMP DL,80
	JC CURROK
	DEC DL
CURROK:
	MOV AH,2
	MOV AL,0
	MOV BX,0
	INT 10H
	RET

     ENDIF

;---------------------------------------------------
;GET KEYBOARD STATUS
GETKEYST:
	CMP B,[BYTERDYFLG],00H	;DO WE ALREADY HAVE A CHAR?
	JNZ HAVENOW

	MOV AH,06H	;DON'T HAVE ONE, SO CHECK ON
	MOV DL,0FFH	;INPUT STATUS
	INT 21H
	CMP AL,00H	;DID WE PICK UP A CHAR?
	JZ NOTREADY
	AND AL,7FH	;STRIP PARITY
	MOV [BYTEIN],AL	;AND SAVE THE CHAR
	MOV [GETIN],AL	;

HAVENOW:
	MOV AL,0FFH
	MOV [BYTERDYFLG],AL	;GOT A CHAR, SO
	MOV [STRDY],AL		;SET FLAGS AND RET
	JMP DKEYSTX
NOTREADY:
	MOV AL,00H	;NO CHAR, SO RET ZERO
DKEYSTX:
	OR AL,AL
	RET


;GET RS232 STATUS
GETCOMMST:
	CMP B,[COMBRDYFLG],00H	;CHECK IF WE ALREADY HAVE A CHAR
	JNZ HAVECOMNOW
	MOV AH,03H	;DON'T HAVE ONE, SO CHECK ON IT
	MOV DX,0
	INT 14H		;STANDARD RS232 INT
	AND AL,01H
	JZ NOTREADY	;NOTHING THERE
	MOV AH,02H
	MOV DX,0	;GOT A CHAR OUT THERE, SO GO GET IT
	INT 14H
	AND AL,7FH		;STRIP THE PARITY
	MOV [COMBYTEIN],AL	;AND SAVE IT
	MOV [GETIN],AL
	JMP HAVECOMNOW

HAVECOMNOW:
	MOV AL,0FFH		;GOT A CHAR
	MOV [COMBRDYFLG],AL	;SO SET THE FLAGS
	MOV [STRDY],AL
	OR AL,AL
	RET


;GET A CHAR FROM THE CONSOLE
GETKEYIN:
	CMP B,[BYTERDYFLG],00H	;CHECK IF WE ALREADY HAVE ONE
	JNZ GOTALREADY
	CALL GETKEYST		;NO CHAR AVAIL, SO WAIT FOR IT
	JMP GETKEYIN

GOTALREADY:
	MOV B,[STRDY],00H
	MOV B,[BYTERDYFLG],00H	;GOT A CHAR, SO CLEAR THE FLAGS
	MOV AL,[BYTEIN]		;AND RET THE CHAR
	RET


;GET A CHAR FROM THE RS232 PORT
GETCOMMIN:
	CMP B,[COMBRDYFLG],00H	;CHECK IF WE ALREADY HAVE ONE
	JNZ GOTCOMMNOW
	CALL GETCOMMST		;NO CHAR AVAIL, SO WAIT FOR IT
	JMP GETCOMMIN
GOTCOMMNOW:
	MOV B,[STRDY],00H
	MOV B,[COMBRDYFLG],00H	;GOT A CHAR, SO CLEAR THE FLAGS
	MOV AL,[COMBYTEIN]	;AND RET THE CHAR
	RET


;OUTPUT A CHAR TO THE RS232 PORT
GETCOMMOUT:
	MOV AH,01H
	MOV AL,DL	;STANDARD RS232 INT 14
	MOV DX,0
	INT 14H
	RET

;SEND CHAR TO THE PRINTER 
GETPRINOUT:
	MOV AH,05H	;STANDARD PRINTER OUTPUT
	INT 21H
	RET


;STANDARD CONSOLE OUTPUT
BCONOT:
	AND DL,7FH		;STRIP PARITY
	CMP B,[CONFLAG],00H	;CHECK ON THE TYPE OF OUTPUT
	JZ NOTBDCONOT
	CALL DIRCONCHK	;ADAPT CHAR FOR CPM TYPE OUTPUT
	CMP AL,00H
	JNZ NOTBDCONOT
	CALL STOPCHECK	;CHECK FOR ^S STOP
	RET

NOTBDCONOT:
	CALL DOBCONOT	;DO THE OUTPUT
	CALL STOPCHECK	;CHECK FOR ^S STOP
	RET

DOBCONOT:
	MOV AL,[03H]	;CHECK FOR IO REDIRECT
	AND AL,03H
	JZ SCREEN	;CONSOLE SCREEN
	CMP AL,01H
	JZ COMMOUT	;RS232 OUTPUT
	CMP AL,02H
	JZ PRINOUT	;PRINTER OUTPUT
	CALL COMMOUT
	JMP SCREEN	;RS232 AND CONSOLE OUTPUT


;STANDARD CONSOLE INPUT
BCONIN:
	CMP B,[CONFLAG],00H	;CHECK ON TYPE OF INPUT
	JZ DOBCONIN	;GO GET THE CHAR
	CALL DOBCONIN	;ELSE DO ADAPTIVE INPUT
	JMP DIRCINCHK	;ADAPT INPUT TO CPM STYLE 

DOBCONIN:
	MOV AL,[03H]	;CHECK FOR IO BYTE REDIRECT
	AND AL,03H
	JZ KEYIN	;CONSOLE INPUT
	CMP AL,01H
	JZ COMMIN	;RS232 INPUT
	CMP AL,02H
	JZ COMMIN	;RS232 INPUT
	CALL KEYST
	OR AL,AL
	JNZ KEYIN	;CONSOLE INPUT
	JMP COMMIN	;OR RS232 INPUT


;GET STANDARD CONSOLE STATUS
BCONST:
	MOV AL,[03H]	;CHECK FOR IO BYTE REDIRECT
	AND AL,03H
	JZ KEYST	;CONSOLE STATUS
	CMP AL,01H
	JZ COMMST	;RS232 STATUS
	CMP AL,02H
	JZ COMMST	;RS232 STATUS
	CALL COMMST
	MOV AH,AL
	PUSH AX
	CALL KEYST	;CONSOLE OR RS232 STATUS
	POP AX
	OR AL,AH
	RET

KEYST:
	JMP GETKEYST
COMMST:
	JMP GETCOMMST
KEYIN:
	JMP GETKEYIN
COMMIN:
	JMP GETCOMMIN
SCREEN:
	JMP GETSCREEN
COMMOUT:
	JMP GETCOMMOUT
PRINOUT:
	JMP GETPRINOUT


;GET A CHARACTER FROM STANDARD READER INPUT
BRDIN:
	MOV AL,[03H]	;CHECK ON IO BYTE REDIRECT
	AND AL,0CH
	JZ KEYIN	;CONSOLE INPUT
	CMP AL,04H
	JZ COMMIN	;RS232 INPUT
	CMP AL,08H
	JZ KEYIN	;CONSOLE INPUT
	JMP COMMIN	;RS232 INPUT

;OUTPUT A CHARACTER VIA THE STANDARD PUNCH OUTPUT
BPUNOUT:
	MOV AL,[03H]	;CHECK ON IO BYTE REDIRECT
	AND AL,30H
	JZ SCREEN	;CONSOLE OUTPUT
	CMP AL,10H
	JZ COMMOUT	;RS232 OUTPUT
	CMP AL,20H
	JZ PRINOUT	;PRINTER OUTPUT
	JMP COMMOUT	;RS232 OUTPUT


;OUTPUT A CHARACTER VIA THE STANDARD PRINTER OUTPUT
BLISTOUT:
	MOV AL,[03H]	;CHECK ON IO BYTE REDIRECT
	AND AL,0C0H
	JZ SCREEN	;CONSOLE OUTPUT
	CMP AL,040H
	JZ COMMOUT	;RS232 OUTPUT
	CMP AL,080H
	JZ PRINOUT	;PRINTER OUTPUT
	JMP COMMOUT	;RS232 OUTPUT


;CHECK INPUTED CHAR TO SEE IF IT WAS ^S AND WAIT IF IT WAS
STOPCHECK:
	PUSH DX
	CALL BCONST	;CHECK IF A CHAR IS OUT THERE
	CMP B,[STRDY],00H
	JZ NOSTPCHR
	CMP B,[GETIN],'S'-40H	;WAS IT A ^S?
	JNZ NOSTPCHR
	CALL BCONIN	;CLEAR THIS CHR
	CALL BCONIN	;WAIT FOR NEXT ONE
NOSTPCHR:
	POP DX
	RET

;ADAPT CHAR TO OUTPUT TO FIT CPM OUTPUT CHARACTERISTICS 
DIRCONCHK:
	CMP DL,00H
	JZ IGNOREOUT	;IGNORE NULLS
	CMP DL,7FH
	JZ IGNOREOUT	;AND RUBOUTS
	CMP DL,'I'-40H
	JNZ CONOUTOK	;EXPAND TABS

TABOUTLP:
	MOV DL,' '
	CALL BCONOT
	MOV AL,[SPCOUNT]	;EXPAND TABS TO LAST +8 
	AND AL,07H
	JNZ TABOUTLP

IGNOREOUT:
	MOV AL,00H	;RETURN NULL FOR IGNORE
	RET

CONOUTOK:
	CMP DL,0DH	;IF NOT CR, THEN ADJUST CHAR COUNTER
	JNZ INCSPC1
	MOV B,[SPCOUNT],00H	;OTHERWISE CLEAR THE COUNTER
	JMP CONTSPC

DECSPC:
	CMP B,[SPCOUNT],00H	;DON'T BACKUP IF AT ZERO
	JZ CONTSPC
	DEC B,[SPCOUNT]		;OTHERWISE BACKUP 
	JMP CONTSPC

INCSPC1:
	CMP DL,'H'-40H		;IF ^H BACKUP A CHAR
	JZ DECSPC
	CMP DL,' '		;IF NOT PRINTABLE CHAR DON'T COUNT THE CHAR
	JC CONTSPC
	CMP DL,7FH		;NOR IF IT IS A RUBOUT
	JZ CONTSPC
	INC B,[SPCOUNT]		;OTHERWISE COUNT IT
CONTSPC:
	CMP B,[PRTFLAG],0FFH	;PRINTER ON?
	JNZ NOPRTOUT
	PUSH DX
	CALL BLISTOUT		;IF SO SEND THE CHAR OUT TO THE PRINTER TOO
	POP DX
NOPRTOUT:
	MOV AL,0FFH		;OTHERWISE WE'RE DONE
	RET

;ADAPT CHAR INPUTED TO FIT CPM INPUT CHARACTERISTICS 
DIRCINCHK:
	CMP AL,7FH	;IGNORE RUBOUTS
	JZ NOCINECHO
	CMP AL,'H'-40H	;^H IS OK
	JZ CINECHO
	CMP AL,'I'-40H	;TABS ARE OK
	JZ CINECHO
	CMP AL,'J'-40H	;LF IS OK
	JZ CINECHO
	CMP AL,'M'-40H	;CR IS OK
	JZ CINECHO
	CMP AL,' '
	JC NOCINECHO	;ALL OTHER CONTROL CHARS ARE NOT ECHOED
CINECHO:
	PUSH DX
	PUSH AX
	MOV DL,AL
	CALL BCONOT	;ECHO THE INPUTED CHAR
	POP AX
	POP DX
NOCINECHO:
	RET

	DB 'Oh dear, you are looking around in here? Sigh...',CR,LF 
	DB 'That does not bode too well I would think. It is beginning',CR,LF
	DB 'to look to me like you are intending to steal my code.',CR,LF
	DB 'Sigh... Now I am getting depressed...',CR,LF 
	DB 'Come on, it is only $40 for the source code.',CR,LF
	DB 'Is that too much to ask?',CR,LF  
	DB 0FFH
;
;
; SIMULATE CP/M JMP TABLE CALLS WITH MS DOS CALLS
;

JCONST:
	PUSH ES
	PUSH DS
	PUSH SI
	PUSH DI
	PUSH BP
	PUSH DX
	PUSH CX
	PUSH BX 

	MOV B,[CONFLAG],00H	;CONSOLE STATUS JMP
	CALL BCONST
	OR AL,AL
	LAHF

	POP BX
	POP CX
	POP DX
	POP BP
	POP DI
	POP SI
	POP DS
	POP ES
	JMP DONE

JCONIN:
	PUSH ES
	PUSH DS
	PUSH SI
	PUSH DI
	PUSH BP
	PUSH DX
	PUSH CX
	PUSH BX 

	MOV B,[CONFLAG],00H	;CONSOLE INPUT JMP
	CALL BCONIN
	LAHF

	POP BX
	POP CX
	POP DX
	POP BP
	POP DI
	POP SI
	POP DS
	POP ES
	JMP DONE

JCONOUT:
	PUSH ES
	PUSH DS
	PUSH SI
	PUSH DI
	PUSH BP
	PUSH DX
	PUSH CX
	PUSH BX 

	MOV B,[CONFLAG],00H	;CONSOLE OUTPUT JMP
	MOV DL,CL
	CALL BCONOT
	LAHF

	POP BX
	POP CX
	POP DX
	POP BP
	POP DI
	POP SI
	POP DS
	POP ES
	JMP DONE

JLIST:
	PUSH ES
	PUSH DS
	PUSH SI
	PUSH DI
	PUSH BP
	PUSH DX
	PUSH CX
	PUSH BX 

	MOV DL,CL
	MOV B,[CONFLAG],00H	;PRINTER OUTPUT JMP
	CALL BLISTOUT
	LAHF

	POP BX
	POP CX
	POP DX
	POP BP
	POP DI
	POP SI
	POP DS
	POP ES
	JMP DONE

JPUNCH:
	PUSH ES
	PUSH DS
	PUSH SI
	PUSH DI
	PUSH BP
	PUSH DX
	PUSH CX
	PUSH BX 

	MOV DL,CL
	MOV B,[CONFLAG],00H	;PUNCH OUTPUT JMP
	CALL BPUNOUT
	LAHF

	POP BX
	POP CX
	POP DX
	POP BP
	POP DI
	POP SI
	POP DS
	POP ES
	JMP DONE

JREADER:
	PUSH ES
	PUSH DS
	PUSH SI
	PUSH DI
	PUSH BP
	PUSH DX
	PUSH CX
	PUSH BX 

	MOV B,[CONFLAG],00H	;READER INPUT JMP
	CALL BRDIN
	LAHF

	POP BX
	POP CX
	POP DX
	POP BP
	POP DI
	POP SI
	POP DS
	POP ES
	JMP DONE

JHOME:
	PUSH ES
	PUSH DS
	PUSH SI
	PUSH DI
	PUSH BP
	PUSH DX
	PUSH CX
	PUSH BX 

	MOV AH,00H	;MOVE DISK DRIVE HEAD TO HOME POS
	INT 13H		;(TRACK 0)
	LAHF

	POP BX
	POP CX
	POP DX
	POP BP
	POP DI
	POP SI
	POP DS
	POP ES
	JMP DONE

JSELDSK:
	PUSH ES
	PUSH DS
	PUSH SI
	PUSH BP
	PUSH DX
	PUSH CX
	PUSH BX 

	MOV DL,CL
	MOV [DISKREG],DL	;SELECT A DIFFERENT DRIVE
	MOV AH,0EH
	PUSH DX
	INT 21H
	POP DX
	CMP DL,AL
	MOV BX,DPARM	;RETURN PARM TABLE POINTER
	JC DSKSELEX
DISKERROR:
	STC
	MOV BX,0000H	;INVALID SELECT, SO INDICATE ERROR
	JMP DSKSELEX2

DSKSELEX:
	OR AL,AL
DSKSELEX2:
	MOV DI,BX
	LAHF
	POP BX
	POP CX
	POP DX
	POP BP
	POP SI
	POP DS
	POP ES
	JMP DONE

JSETTRK:
	JMP BIOSERROR	;TRACK SELECT IS NOT IMPLIMENTED

JSETSEC:
	JMP BIOSERROR	;SECTOR SELECT IS NO IMPLIMENTED

JSETDMA:
	PUSH ES
	PUSH DS
	PUSH SI
	PUSH DI
	PUSH BP
	PUSH DX
	PUSH CX
	PUSH BX 

	MOV [DMA],CX	;SET DMA TRANSFER ADDRESS
	MOV DX,CX
	MOV AH,1AH
	INT 21H

	POP BX
	POP CX
	POP DX
	POP BP
	POP DI
	POP SI
	POP DS
	POP ES
	JMP DONE

JREAD:
	JMP BIOSERROR	;DISK READ IS NOT IMPLIMENTED
JWRITE:
	JMP BIOSERROR	;DISK WRITE IS NOT IMPLIMENTED

JLISTST:
	OR AL,255	;PRINTER STATUS IS ALWAYS READY
	LAHF
	JMP DONE

JSECTRAN:
	MOV DI,CX	;SECTOR TRANSLATE IS DIRECT LOGICAL TO PHYSICAL
	INC DI
	JMP DONE

DOBDOS:
	JMP RBDOS	;GO SERVICE THE BDOS REQUEST

;THIS DIRECTS THE JMP TABLE CALLS TO THE PROPER ROUTINES
RBIOS:
	CMP AL,7FH	;WAS IT A BDOS REQUEST?
	JZ DOBDOS	;YUP, GO DO IT
	PUSH BX
	MOV BH,0
	MOV BL,AL
	CMP BL,MAXBIOSFUN	;DON'T GO PAST MAX ALLOWED CALL 
	MOV DI,BX
	POP BX
	JNC BIOSERROR
	SAL DI
	SEG CS
	JMP [BIOSFUNTABLE+DI]	;GET FUNCTION ADDRESS FROM TABLE

MAXBIOSFUN	EQU 18

BIOSFUNTABLE:
	DW BIOSERROR	;0 IS INVALID
	DW USERBOOT	;1 COLD BOOT
	DW USERBOOT	;2 WARM BOOT
	DW JCONST	;3 CONSOLE STATUS
	DW JCONIN	;4 CONSOLE INPUT
	DW JCONOUT	;5 CONSOLE OUTPUT
	DW JLIST	;6 PRINTER OUTPUT
	DW JPUNCH	;7 PUNCH OUTPUT
	DW JREADER	;8 READER INPUT
	DW JHOME	;9 HOME DISK DRIVE
	DW JSELDSK	;10 SELECT DISK DRIVE
	DW JSETTRK	;11 SELECT TRACK
	DW JSETSEC	;12 SELECT SECTOR
	DW SETDMA	;13 SET DMA TRANSFER ADDRESS
	DW JREAD	;14 READ FROM DISK
	DW JWRITE	;15 WRITE TO DISK
	DW JLISTST 	;16 PRINTER STATUS
	DW JSECTRAN	;17 SECTOR TRANSLATE



;IF THERE IS AN ERROR IN THE BIOS SELECTION WE COME HERE
BIOSERROR:
	SEG DS
	MOV SI,[BP]	;SAVE CPM STACK ADDRESS
	PUSH AX
	PUSH DX
	MOV DL,CR	;PRINT A CR
	MOV AH,2
	INT 21H
	MOV DL,LF	;FOLLOWED BY LF
	MOV AH,2
	INT 21H
	MOV DL,BELL	;FOLLOWED BY BELL
	MOV AH,2
	INT 21H
	POP DX
	POP AX
	CALL PRTBYTE	;THEN PRINT THE BIOS FUNCTION NUMBER 

	PUSH DS
	PUSH BX
	PUSH AX
	MOV BX,CS
	MOV DS,BX
	MOV DX,BIOSERRMSG	;FOLLOWED BY THE ERROR MSG
	MOV AH,09H
	INT 21H
	POP AX
	POP BX
	POP DS

	CALL HEADER	;PRINT THE ERROR ID HEADER INFO
	CALL DEBUG	;FOLLOWED BY THE ERROR INFORMATION
	JMP USERBOOT	;AND RETURN TO MSDOS

BIOSERRMSG:
	DB ' <- Invalid BIOS call ***',BELL,CR,LF,'$'
	DB 'If you bought the source code you would know why',CR,LF 
	DB 'that message gets displayed.',CR,LF,0FFH

;*******************************************************
;THIS IS WHERE WE COME IF WE GET A BDOS FUNCTION ERROR
BDOSFUNERROR:
	SEG DS
	MOV SI,[BP]	;SAVE THE CPM STACK
	PUSH AX
	PUSH DX
	MOV DL,CR	;PRINT A CR
	MOV AH,2
	INT 21H
	MOV DL,LF	;FOLLOWED BY A LF
	MOV AH,2
	INT 21H
	MOV DL,BELL	;FOLLOWED BY A BELL
	MOV AH,2
	INT 21H
	POP BX
	POP AX
	PUSH AX
	MOV AL,AH
	CALL PRTBYTE	;THEN PRINT THE BDOS FUNCTION NUMBER OF THE ERROR
	POP AX

	PUSH DS
	PUSH BX
	PUSH AX
	MOV BX,CS
	MOV DS,BX
	MOV DX,FUNERRMSG	;FOLLOWED BY THE ERROR MSG
	MOV AH,09H
	INT 21H
	POP AX
	POP BX
	POP DS

	CALL HEADER	;PRINT THE DEBUG HEADER
	CALL DEBUG	;FOLLOWED BY THE ERROR MSG
	JMP USERBOOT	;AND GO BACK TO MSDOS 

FUNERRMSG:
	DB ' <- Invalid BDOS function call ***',BELL,CR,LF,'$'
	DB 'Remember, for 40 bucks you get the manual too.',CR,LF,'$'

BDOSFUNEX:
	JMP BDOSFUNERROR

;************************************************
;WE COME HERE TO PROCESS THE CPM BDOS FUNCTION CALLS
RBDOS: 
	PUSH SI
	PUSH BP
	PUSH ES

	PUSH DS
	MOV AH,CL

	MOV B,[DMAFLAG],00H
	MOV AL,00H
	MOV BH,0
	MOV BL,AH
	CMP BL,MAXBDOSFUN	;DON'T ALLOW CALLS OUTSIDE RANGE
	JNC BDOSFUNEX
	SAL BX
	SEG CS
	JMP [BDOSFUNTABLE+BX]	;AND GO DO THE CALL

NORMBDOS:
	INT 21H	;IF NO MOD TO THE CALL IS NEEDED WE JUST GO DIRECT TO DOS

;WHEN THE CALL IS DONE WE COME HERE TO CLEAN THINGS UP
BDOSDONE:
	CMP B,[DMAFLAG],00H	;FIX DMA?
	JZ NODMAMOD

	PUSH AX
	PUSH BX
	PUSH CX
	PUSH DX
	MOV DX,[DMA]	;IF NEEDED WE READJUST THE DMA HERE
	MOV AH,1AH
	INT 21H
	POP DX
	POP CX
	POP BX 
	POP AX

NODMAMOD:
	POP DS
	MOV DI,BX	;OTHERWISE WE FIXUP THE REGISTERS  
	OR AL,AL	;AND FLAGS
	LAHF
	POP ES
	POP BP
	POP SI
	JMP DONE	;AND GET THE NEXT INSTRUCTION


;THIS TABLE CORESPONDS TO THE CPM FUNCTION SELECT
BDOSFUNTABLE:
	DW USERBOOT	;0
	DW BDOSCONIN	;1
	DW BDOSCONOT	;2
	DW BDOSRDIN	;3
	DW BDOSPNOUT	;4
	DW BDOSLOUT	;5
	DW BDOSDIO	;6
	DW GETIO	;7
	DW PUTIO	;8
	DW BDOSPSOUT	;9
	DW BDOSRBUF	;10
	DW BDOSCONST	;11
	DW BVERSION	;12
	DW RESETSYS	;13
	DW SELDISK	;14
	DW OPENFILE	;15
	DW CLOSEFILE	;16
	DW SRCHFST	;17
	DW SRCHNXT	;18
	DW NORMBDOS	;19
	DW READSEQ	;20
	DW WRITESEQ	;21
	DW MAKEFILE	;22
	DW NORMBDOS	;23
	DW GETDLOG	;24
	DW NORMBDOS	;25
	DW SBDOSDMA	;26
	DW GETALLOC	;27
	DW NULLBDOS	;28
	DW NULLBDOS	;29
	DW NULLBDOS	;30
	DW GETPARM	;31
	DW NULLBDOS	;32
	DW RANREAD	;33
	DW RANWRITE	;34
	DW GETFSIZE	;35
	DW NORMBDOS	;36
	DW DRVRESET	;37
	DW BDOSFUNERROR	;38
	DW BDOSFUNERROR	;39
	DW RANZWRITE	;40
	DW BDOSFUNERROR	;41

MAXBDOSFUN: EQU 41

;OPEN A FILE
OPENFILE:
	PUSH BX
	MOV BX,DX
	MOV [BX+14],00H
	POP BX
	JMP NORMBDOS

;CLOSE A FILE
CLOSEFILE:
	JMP NORMBDOS

;GET THE FILE SIZE
GETFSIZE:
	JMP NORMBDOS

;SEQUENCIAL READ OF A RECORD
READSEQ:
	MOV B,[DMAFLAG],0FFH
	INT 21H
	CMP AL,03H
	JNZ READSEQX
	MOV AL,00H
READSEQX:
	JMP BDOSDONE

;RANDOM READ OF A RECORD
RANREAD:
	MOV B,[DMAFLAG],0FFH
	JMP NORMBDOS

;SEQUENCIAL WRITE OF A RECORD
WRITESEQ:
	MOV B,[DMAFLAG],0FFH
	PUSH BX
	MOV BX,DX
	MOV B,[BX+13],00H	;PATCH FCB
	POP BX
	JMP NORMBDOS

;RANDOM WRITE OF A RECORD WITH ZERO FILL
RANZWRITE:
	MOV B,[DMAFLAG],0FFH
	JMP NORMBDOS

;RANDOM WRITE OF A RECORD WITH 
RANWRITE:
	MOV B,[DMAFLAG],0FFH
	JMP NORMBDOS

;MAKE A FILE
MAKEFILE:
	PUSH BX
	MOV BX,DX
	MOV B,[BX+14],00H
	POP BX
	JMP NORMBDOS


;RESET A DRIVE
DRVRESET:
	MOV AH,0DH
	JMP NORMBDOS

;RESET THE SYSTEM
RESETSYS:
	MOV [DRVLOG],01H
	JMP NORMBDOS

;SET UP THE DMA
SBDOSDMA:
	MOV [DMA],DX
	JMP NORMBDOS

;DON'T DO ANYTHING (UNMATCHING CPM/MSDOS FUNCTION)
NULLBDOS:
	MOV AL,00H
	MOV BX,0000H
	JMP BDOSDONE

;RETURN CPM VERSION NUMBER
BVERSION:
	MOV AL,22H
	MOV BX,0022H
	JMP BDOSDONE

;SEARCH FOR FIRST DIRECTORY ENTRY 
SRCHFST:
	PUSH SI
	PUSH DI
	MOV SI,DX
	MOV DI,TEMPFCB	;WE USE TEMP FCB BECAUSE MSDOS
	MOV CX,48	;SMASHES THE FCB WHEN IT DOES THIS
	REP
	MOVB
	POP DI
	POP SI

	MOV DI,DX
	CMP B,[DI+0CH],'?'	;EXTENT #
	JZ SRCHFSTOK
	CMP B,[DI+0CH],00H
	JZ SRCHFSTOK
	MOV AL,0FFH
	JMP SRCHFSTEND

SRCHFSTOK:
	PUSH DX
	MOV DX,TEMPFCB	;FOUND THE FILE, SO FIX THINGS UP
	PUSH DS
	INT 21H
	POP DS
	CALL FIXDMA	;ADJUST THINGS AT THE DMA
	POP DX
SRCHFSTEND:
	JMP BDOSDONE


;SEARCH FOR THE NEXT DIRECTORY ENTRY
SRCHNXT:
	PUSH DX 
	MOV DX,TEMPFCB	;WE ASSUME THAT THE TEMPFCB WAS ALREADY SETUP

	MOV DI,DX
	CMP B,[DI+0CH],'?'	;EXTENT #
	JZ SRCHNXTOK
	CMP B,[DI+0CH],00H
	JZ SRCHNXTOK
	MOV AL,0FFH
	JMP SRCHNXTEND

SRCHNXTOK:
	PUSH DS
	INT 21H
	POP DS
	CALL FIXDMA	;ADJUST THINGS AT THE DMA
	POP DX
SRCHNXTEND:
	JMP BDOSDONE

;THIS ADJUSTS THE DMA TO FIT THE EXPECTED CPM FORMAT
FIXDMA:
	PUSH DX
	PUSH BX
	PUSH AX
	MOV BX,[DMA]
	MOV AH,0FH
	MOV DX,BX
	INT 21H
	MOV BX,[DMA]
	MOV B,[BX],00H		;SET USR AS 00 
	MOV B,[BX+0CH],00H	;CLEAR EXT

	MOV AX,[BX+18]		;POKE FILE SIZE INFO
	CMP AX,0000H
	JNZ FULLSIZE
	MOV AX,[BX+16]
	CMP AX,03FFFH
	JNC FULLSIZE
	ADD AX,AX
	CMP AL,00H
	JZ FIXDMADONE
	INC AH

FIXDMADONE:
	MOV [BX+15],AH

	PUSH BX
	ADD BX,10H
	MOV AL,16
FIXDMABLKLP:
	MOV B,[BX],00H	;MARK ALL GROUPS IN DMA AS ZERO
	INC BX
	DEC AL
	JNZ FIXDMABLKLP
	POP BX 

	POP AX
	POP BX
	POP DX
	RET

FULLSIZE:
	MOV AH,80H
	JMP FIXDMADONE

;GET THE IO BYTE
GETIO:
	MOV AL,[03H]
	MOV BL,AL
	MOV BH,00H
	JMP BDOSDONE

;WRITE A NEW IO BYTE
PUTIO:
	MOV AL,AH
	MOV BL,AL
	MOV BH,00H
	MOV [03H],AL
	JMP BDOSDONE

;READ IN THE DRIVE SELECT REGISTER
GETDLOG:
	MOV BX,[DRVLOG]
	MOV AL,BL
	JMP BDOSDONE

;SELECT A NEW DEFAULT DRIVE
SELDISK:
	MOV AL,DL
	MOV BX,01H
DLOGLOOP:
	DEC AL
	JC DLOGDN
	ROL BX
	JMP DLOGLOOP
DLOGDN:
	OR BX,[DRVLOG]
	MOV [DRVLOG],BX
	JMP NORMBDOS

;GET THE ADDRESS OF THE ALLOCATION TABLE
GETALLOC:
	MOV BX,ALLOC
	MOV AL,BL
	JMP BDOSDONE

;GET ADDRESS OF THE DRIVE PARM TABLE
GETPARM:
	MOV BX,DPARM
	MOV AL,BL
	JMP BDOSDONE

;INPUT A CHARACTER FROM THE CONSOLE
BDOSCONIN:
	MOV B,[CONFLAG],0FFH
	CALL BCONIN
	JMP BDOSDONE

;OUTPUT A CHARACTER TO THE CONSOLE
BDOSCONOT:
	MOV B,[CONFLAG],0FFH
	CALL BCONOT
	JMP BDOSDONE

;INPUT A CHARACTER FROM THE READER
BDOSRDIN:
	MOV B,[CONFLAG],0FFH
	CALL BRDIN
	JMP BDOSDONE

;OUTPUT A CHARACTER TO THE PUNCH
BDOSPNOUT:
	MOV B,[CONFLAG],0FFH
	CALL BPUNOUT
	JMP BDOSDONE

;OUTPUT A CHARACTER TO THE PRINTER
BDOSLOUT:
	MOV B,[CONFLAG],0FFH
	CALL BLISTOUT
	JMP BDOSDONE

;GET CONSOLE INPUT STATUS
BDOSCONST:
	MOV B,[CONFLAG],0FFH
	CALL BCONST
	JMP BDOSDONE

;DIRECT CONSOLE IO (FUNCTION 6)
BDOSDIO:
	MOV B,[CONFLAG],00H
	CMP DL,0FFH	;IF 0FFH CHAR DO INPUT
	JZ BDOSDST
	CALL BCONOT	;ELSE DO OUTPUT
	JMP BDOSDONE
BDOSDST:
	CALL BCONST	;IF INPUT, CHECK ON STATUS
	CMP AL,00H
	JNZ BDOSDOUT	;IF A CHAR IS OUT THERE
	JMP BDOSDONE
BDOSDOUT:
	CALL BCONIN	;GET IT
	JMP BDOSDONE

;PRINT A STRING TERMINATED WITH "$" 
BDOSPSOUT:
	MOV B,[CONFLAG],0FFH
	MOV BX,DX
BPSOUTLP:
	MOV DL,[BX]
	CMP DL,'$'	;KEEP PRINTING UNTIL "$"
	JZ BPSOUTDN
	PUSH BX
	CALL BCONOT	;OUTPUT THE CHAR
	POP BX
	INC BX
	JMP BPSOUTLP
BPSOUTDN:
	JMP BDOSDONE

;**********************************************
;THE FOLLOWING SIMULATES THE CPM BUFFERED KEYBOARD INPUT
FLIPPRT:
	XOR B,[PRTFLAG],0FFH	;TOGGLE PRINTER OUTPUT
	JMP INBUFA

INBUFR:
	CALL RBBCRLF		;TERMINATE WITH CR/LF
	JMP BDOSDONE		;AND RETURN TO NORMAL PROGRAM FLOW

IPCRLF:
	CALL RBBCRLF
	MOV B,[BEGBUFF],00H	;OUTPUT A CRLF AND START OVER
	JMP INBUFA

BACKSPACE:
	CALL BACKUPCHR		;BACKUP ONE CHAR
	JMP INBUFA

BDOSRBUF:
INBUFF:
	MOV BX,DX
	INC BX
	MOV B,[BX],00H		;INIT COUNTERS AND POINTERS
	MOV AL,[SPCOUNT]
	MOV [BEGBUFF],AL
	MOV B,[DELCOUNT],00H

INBUFA:
	PUSH DX
	PUSH BX
	PUSH CX
	MOV B,[CONFLAG],00H
	CALL BCONIN		;GET A CHAR
	POP CX
	POP BX
	POP DX
	AND AL,7FH	;STRIP PARITY
	CMP AL,0AH
	JZ INBUFR	;LINE FEED IS SAME AS CR
	CMP AL,0DH
	JZ INBUFR	;IF CR, THEN WE'RE DONE
	CMP AL,7FH
	JZ IDELETE	;ECHO AND DELETE ON RUBOUT 
	CMP AL,08H
	JZ BACKSPACE	;ERASE AND DELETE ON BS
	CMP AL,'U'-40H
	JZ INBUFU	;CANCEL AND RESTART LINE ON ^U
	CMP AL,'X'-40H
	JZ INBUFX	;CANCEL CURRENT LINE ON ^X 
	CMP AL,'R'-40H
	JZ RETYPE	;RETYPE CURRENT LINE ON ^R
	CMP AL,'E'-40H
	JZ IPCRLF	;NEW LINE ON ^E
	CMP AL,'P'-40H
	JZ FLIPPRT	;TOGGLE PRINTER ON ^P
	JMP SAVEBCHR	;NOT A SPECIAL CHAR, SO SAVE IT TO THE BUFFER 

;PRINT A "#" AT THE END OF THE CURRENT LINE, DO A CR/LF
;AND PREP FOR INPUT ON A NEW LINE
INBUFU:
	MOV AL,'#'
	CALL RBBCONOT
	CALL RBBCRLF
	CALL SETBEGBUFF
	MOV B,[BX],00H
	MOV B,[DELCOUNT],00H
	JMP INBUFA

;BACKUP UNTIL AT THE BEGINNING OF THE CURRENT LINE
;ERASEING CHARACTERS AS WE GO.
INBUFX:
	MOV AL,[SPCOUNT]
	CMP AL,[BEGBUFF]
	JZ INBUFXDN
	CALL ERASECHAR
	JMP INBUFX
INBUFXDN:
	MOV B,[BX],00H
	JMP INBUFA

;PRINT A "#" AT THE END OF THE CURRENT LINE, PRINT A CR/LF
;AND PREPARE FOR A NEW INPUT LINE
RETYPE:
	MOV B,[DELCOUNT],00H
	MOV AL,'#'
	CALL RBBCONOT
	CALL RBBCRLF
	CALL SETBEGBUFF
	CMP B,[BX],00H
	JZ RETYPEDN
	CALL TYPEITBACK
RETYPEDN:
	JMP INBUFA

;REMOVE A CHARACTER FROM THE BUFFER AND ECHO IT ON THE SCREEN
IDELETE:
	CMP B,[BX],00H
	JZ NODELETE
	CALL RDBUFCHR
	PUSH AX
	CALL RBBCONOT
	POP AX
	DEC B,[BX]

	CMP AL,'I'-40H	;TAB BACKUP? 
	JZ TABBACKUP
	CMP AL,' '
	JNC NOTCTRLDEL
	ADD B,[DELCOUNT],2	;CONTROL CHAR BACKUP
NOTCTRLDEL:
	ADD B,[DELCOUNT],2
	JMP INBUFA

TABBACKUP:
	MOV AL,[SPCOUNT]	;BACKUP OVER A TAB
	AND AL,07H
	JNZ NOTFULLTAB
	MOV AL,08H
NOTFULLTAB:
	ADD [DELCOUNT],AL 
	
NODELETE:
	JMP INBUFA

BACKUPCHR:
	CMP B,[BX],00H
	JZ NOBACKUP
	CMP B,[SPCOUNT],00H
	JZ NOBACKUP

DELERALOOP:
	CMP B,[DELCOUNT],00H
	JZ DELBACKDN
	CALL ERASECHAR
	DEC B,[DELCOUNT]
	JNZ DELERALOOP

DELBACKDN:
	CALL ERASECHAR
	DEC B,[BX]
NOBACKUP:
	RET

ERASECHAR:
	MOV AL,08H
	CALL RBBCONOT
	MOV AL,' '
	CALL RBBCONOT
	MOV AL,08H
	CALL RBBCONOT
	RET

RDBUFCHR:
	PUSH CX
	PUSH BX
	MOV CH,00H
	MOV CL,[BX]
	ADD BX,CX
	MOV AL,[BX]
	POP BX
	POP CX
	RET

TYPEITBACK:
	PUSH CX
	PUSH BX
	MOV CL,[BX]
CTLRLP:
	INC BX
	MOV AL,[BX]
	CALL RBBCONOT
	DEC CL
	JNZ CTLRLP
	POP BX
	POP CX
	RET

;EXPAND TABS
FORWARDTAB:
	PUSH CX
	MOV CL,[SPCOUNT]
	AND CL,07H
FWDTABDN:
	MOV AL,08H
	SUB AL,CL
	DEC AL
	POP CX
	ADD [DELCOUNT],AL
	MOV AL,'I'-40H
	JMP SAVCHRCNT

SAVEBCHR:
	CMP AL,'I'-40H	;IF A TAB, SAVE ^I BUT PRINT SPACES
	JZ FORWARDTAB
	CMP AL,'C'-40H	;IF ^C TREAT SPECIAL
	JNZ SAVCHRCNT
	CMP B,[BX],00H	;IF NOT AT FIRST POS OF LINE
	JNZ SAVCHRCNT	;SAVE IT AS NORMAL
	JMP USERBOOT	;ELSE RETURN TO MSDOS

SAVCHRCNT:
	INC B,[BX]	;SAVE THE CHAR IN THE INPUT BUFFER
	PUSH CX
	PUSH BX
	MOV CH,00H
	MOV CL,[BX]
	ADD BX,CX
	MOV [BX],AL
	POP BX
	CALL RBBCONOT	;HANDLE OUTPUT
	MOV CL,[BX]
	CMP CL,[BX-1]
	POP CX
	JNC IALERT	;IF AT MAX POS STOP AND RING BELL
	JMP INBUFA	;ELSE GET NEXT CHAR

IALERT:
	JMP INBUFR	;OVERFLOW EXIT

RBBCONOT:
	PUSH DX
	PUSH BX
	PUSH CX
	PUSH AX
	CMP AL,0DH	;OUPUT IF IT IS A CR 
	JZ DORBCONOT
	CMP AL,0AH	;SAME IF IT IS LF
	JZ DORBCONOT 
	CMP AL,'I'-40H	;OR TAB
	JZ DORBCONOT
	CMP AL,'H'-40H	;OR BS
	JZ DORBCONOT
	CMP AL,' '	;OR NON-CONTROL CHAR
	JNC DORBCONOT
	PUSH AX
	MOV AL,'^'
	CALL RBBCONOT	;IF CONTROL CHAR PRECEED WITH "^"
	POP AX
	OR AL,40H	;MAKE THE CONTROL CHAR PRINTABLE
DORBCONOT:
	MOV B,[CONFLAG],0FFH
	MOV DL,AL
	CALL BCONOT
	POP AX
	POP CX
	POP BX
	POP DX
	RET

RBBCRLF:
	PUSH AX
	MOV AL,0DH
	CALL RBBCONOT	;PRINT A CR/LF
	MOV AL,0AH
	CALL RBBCONOT
	POP AX
	RET

;FIND THE BEGINING POS OF THE BUFFER LINE ON THE SCREEN
SETBEGBUFF:
	MOV AL,[SPCOUNT]
	CMP AL,[BEGBUFF]
	JZ BEGBUFFDN
	MOV AL,' '
	CALL RBBCONOT
	JMP SETBEGBUFF
BEGBUFFDN:
	RET


MSGLOOP:
	MOV AX,CS	;MAKE SURE IT IS OUR DATA SEGMENT
	MOV DS,AX
	MOV DL,[BX]
	CMP DL,0FFH	;KEEP PRINTING UNTIL FF FOUND 
	JZ MSGDONE
	MOV AH,2
	PUSH BX
	INT 21H
	POP BX
	INC BX
	JMP MSGLOOP
MSGDONE:
	RET

;**********************************************************
;DO ADVERTISMENT MSGS

INTEL:
	MOV BX,INTELMSG	;NO V20/V30 CHIP FOUND SO SHOW ERROR 
	CALL MSGLOOP
	MOV BX,CPMMSG2	;THEN FOLLOW WITH SHORTENED HEADER MSG
	CALL MSGLOOP	;SO THAT EVERYTHING FITS ON THE SCREEN
	JMP USERBOOT

ADVERT:
	MOV BX,CPMMSG	;NO FILENAME, SO PRINT HEADER 
	CALL MSGLOOP
	JMP USERBOOT

;****************************************************
;NOW WE START TO SETUP THE EMULATOR SO WE CAN RUN IT

INIT:
	MOV SP,STACK	;MAKE OUR STACK
	MOV AX,CS	;COPY CS TO DS
	MOV DS,AX

	MOV BX,BLNKSCR	;MAKE SURE SCREEN IS BLANK
	CALL MSGLOOP

     IF NECCHK
	MOV CX,0FFFFH	;TO CHECK FOR NEC V20/V30
	STI		;MOVE A BUNCH OF DATA OUT OF MEM WITH
	REP		;A REPEATED LODS. IF AT THE END CX IS NOT ZERO
	SEG ES		;BECAUSE A CLOCK INTERRUPT OCCURED DURING THE
	LODSB 		;LOOP THEN IT IS AN INTEL PART SO ERROR OUT
	JMP NECEND	;INTEL PARTS DON'T HANDLE THE INSTRUCTION PROPERLY 
NECEND:			;WHEN AN INTERUPT OCCURS WHILE REPEATING. 
	OR CX,CX	;IT GETS TERMINATED EARLY, LEAVING CX NON-ZERO.
	JNZ INTEL
     ENDIF

	CMP B,[5DH],20H	;IF NO FILENAME GIVEN DO ADVERTISMENT
	JZ ADVERT

	MOV SI,2
	MOV BX,[SI]	;GET SIZE OF AVAILABLE MEMORY

; SETUP TEMP FCB
	PUSH BX
	MOV SI,005CH
	MOV DI,FCB
	MOV CX,16
	CLD
	REP
	MOVB
	POP BX

; SET UP SEGMENT REGISTER
	MOV DX,RAMSEGMENT
	MOV CX,DS
	ADD DX,CX
	MOV ES,DX
	MOV [SEG8080],DX	;SAVE 8080 SEGMENT ADDRESS

	MOV AX,DX
	ADD AX,4096	;CHECK IF THERE IS ENOUGH ROOM TO RUN PROGRAM 
	CMP AX,BX
	JC INITSZOK	;SIZE IS OK

	SUB AX,BX
	PUSH AX
	MOV AL,CR
	CALL TYPEITOUT	;ON ERROR SHOW HOW MUCH MEMORY WE'RE SHORT 
	MOV AL,LF
	CALL TYPEITOUT
	MOV AL,BELL
	CALL TYPEITOUT
	MOV AL,'['
	CALL TYPEITOUT
	MOV AL,'-'
	CALL TYPEITOUT
	POP AX
	PUSH AX
	MOV AL,AH
	CALL PRTBYTE
	POP AX
	CALL PRTBYTE
	MOV AL,'0'
	CALL TYPEITOUT

	MOV BX,CS
	MOV DS,BX
	MOV DX,NOMEMMSG	;AND SHOW THE ERROR MESSAGE
	MOV AH,09H
	INT 21H
	JMP USERBOOT		;RE BOOT

INITSZOK:
	JMP INITSETUP

NOMEMMSG:
	DB '] *** Insufficient user memory available to run program ***'
	DB 7,CR,LF,'$' 



;
; SETUP JUMP TABLES OF CPM
;
INITSETUP:
	CLD
	MOV SI,BOOT	;INIT BOTTOM 256 BYTES OF CPM
	MOV DI,0000H
	MOV CX,100H
	REP
	MOVB

	MOV SI,REALBDOS	;INIT BDOS JMP ENTRY
	MOV DI,0EC06H
	MOV CX,4
	REP
	MOVB

	MOV SI,CPMJMP	;INIT BIOS JMP TABLE
	MOV DI,0FA00H
	MOV CX,128
	REP
	MOVB

	MOV SI,CBOOT	;INIT BIOS SERVICE CALLS
	MOV DI,0FB00H
	MOV CX,128
	REP
	MOVB

; SETUP FCB TO READ IN COM FILE

	CMP B,[FCB+9],20H	;WAS THE FCB EXTENT GIVEN?
	JNZ WHOLEFCB

	MOV B,[FCB+9],'C'	;IF NOT, ASSUME "CPM"
	MOV B,[FCB+10],'P'
	MOV B,[FCB+11],'M'

WHOLEFCB:
	MOV SI,FCB
	MOV DI,0FE00H	;TRANSFER THE FCB TO CPM MEMORY
	MOV CX,36
	REP
	MOVB

	SEG CS
	MOV DX,[SEG8080]	;INIT DS TO THE 8080 CODE SEGMENT
	MOV DS,DX

;CREATE DEFDMA COMMAND LINE

	MOV BX,81H
SPDINITLP:
	SEG CS
	MOV AL,[80H]	;IF NO COMMAND LINE GIVEN MAKE IT EMPTY
	CMP AL,00H
	JZ DEFEMPTY
	SEG CS
	MOV AL,[BX]	;ELSE BYPASS FIRST PARM
	CMP AL,20H
	JNZ SPSEARCH	
	INC BX
	SEG CS
	DEC B,[80H]
	JZ DEFEMPTY	;IF NOTHING MORE THEN EMPTY
	JMP SPDINITLP

SPSEARCH:
	SEG CS
	MOV AL,[BX]	;FIND FIRST NON-SPACE POSITION
	CMP AL,20H
	JZ FOUNDSP
	INC BX
	SEG CS
	DEC B,[80H]
	JZ DEFEMPTY	;IF NOTHING MORE
	JMP SPSEARCH

FOUNDSP:
	SEG CS
	MOV AL,[80H]
	MOV [80H],AL
	MOV DI,81H
FILEMOVLP:
	SEG CS
	MOV AL,[BX]	;NOW COPY THE REST OF THE LINE
	MOV [DI],AL	;TO THE CPM COMMAND LINE
	INC BX
	INC DI
	SEG CS
	DEC B,[80H]
 	JNZ FILEMOVLP
	MOV B,[DI],00H
	JMP FILEMOVDN
DEFEMPTY:
	MOV W,[80H],00H

FILEMOVDN:
	MOV B,[BYTERDYFLG],00H	;INIT FLAGS
	MOV B,[COMBRDYFLG],00H
	MOV B,[PRTFLAG],00H
	MOV B,[SPCOUNT],00H
	MOV B,[CURFLAG],00H
	MOV B,[ROWFLAG],00H
	MOV B,[EQFLAG],00H

	MOV W,[DRVLOG],01H	;START WITH DRIVE A ALWAYS LOGGED IN
	MOV AH,19H
	INT 21H			;ALSO LOG IN DEFAULT DRIVE
	INC AL
	MOV BX,1
DRVINTLOOP:
	DEC AL
	JZ DRVINTALLDN
	CLC
	RCL BX
	JMP DRVINTLOOP

DRVINTALLDN:
	OR [DRVLOG],BX

	MOV BX,ALLOC
	MOV CL,00H		;INIT DRV ALLOC TABLE
ALLOCINT:
	MOV B,[BX],00H
	INC BX
	DEC CL
	JNZ ALLOCINT

	MOV AH,30H
	INT 21H
	CMP AL,0
	JNZ SETDP2

	MOV W,[DPARM],32	;SPT	;FOR VERSION 1 MSDOS
	MOV B,[DPARM+2],3	;BSH
	MOV B,[DPARM+3],7	;BLM
	MOV B,[DPARM+4],0	;EXM
	MOV W,[DPARM+5],160	;DSM
	MOV W,[DPARM+7],63	;DRM
	MOV B,[DPARM+9],0C0H	;AL0
	MOV B,[DPARM+10],00H	;AL1
	MOV W,[DPARM+11],16	;CKS
	MOV W,[DPARM+13],2	;OFF
	JMP INITCURPOS

SETDP2:
	MOV W,[DPARM],36	;SPT	;FOR VERSION 2 and up of MSDOS
	MOV B,[DPARM+2],3	;BSH
	MOV B,[DPARM+3],7	;BLM
	MOV B,[DPARM+4],0	;EXM
	MOV W,[DPARM+5],359	;DSM
	MOV W,[DPARM+7],63	;DRM
	MOV B,[DPARM+9],0C0H	;AL0
	MOV B,[DPARM+10],00H	;AL1
	MOV W,[DPARM+11],16	;CKS
	MOV W,[DPARM+13],2	;OFF


INITCURPOS:
	MOV AX,[DPARM+9]	;GET DIR ALLOC
	MOV W,[ALLOC],AX	;SETUP ALLOC TABLE 

;
; OPEN THE COM FILE
;
	MOV BX,0FE01H	;CHECK IF A FILENAME WAS GIVEN
	MOV AL,[BX]
	CMP AL,20H
	JNZ GOTAFILE 
	JMP ADVERT	;COULDN'T FIND A FILENAME, SO GO DO ADVERT



;IF FILNAME GIVEN, THEN TRY TO OPEN IT
GOTAFILE:
	MOV AH,0FH	;OPEN FILE FUNCTION
	MOV DX,0FE00H	;FCB IN HIGH MEMORY
	INT 21H
	OR AL,AL
	JNZ NOFILEFND	;IF COULDN'T FIND IT
	JMP OPENOK

;FILE DOES NOT EXIST MESSAGE
NOFILEFND:
	MOV AL,CR
	CALL TYPEITOUT
	MOV AL,LF
	CALL TYPEITOUT
	MOV AL,BELL
	CALL TYPEITOUT

	MOV DX,0FE00H
	CALL FPRINT	;PRINT THE FILE NAME

	MOV BX,CS
	MOV DS,BX
	MOV DX,NOFILMSG
	MOV AH,09H
	INT 21H
	JMP USERBOOT		;RE BOOT

NOFILMSG:
	DB ' <- That program name does not exist ***',CR,LF,'$'

BIGFILMSG:
	DB ' <- That program is too big to run ***',CR,LF,'$' 

;IF THE FILE WAS TOO BIG WE COME HERE
FILETOOBIG:
	MOV AL,CR
	CALL TYPEITOUT
	MOV AL,LF
	CALL TYPEITOUT
	MOV AL,BELL
	CALL TYPEITOUT

	MOV DX,0FE00H	;SHOW THE FILENAME IN ERROR
	CALL FPRINT

	MOV BX,CS
	MOV DS,BX
	MOV DX,BIGFILMSG
	MOV AH,09H
	INT 21H
	JMP USERBOOT		;RE BOOT

;THE FILE WAS OPENED OK SO WE COME HERE
OPENOK:

;AND READ THE PROGRAM FILE INTO MEMORY 

READFILE:
	MOV BX,[0FE00H+18]
	OR BX,BX
	JNZ FILETOOBIG

	MOV BX,[0FE00H+16]	;GET FILE SIZE TO READ
	CMP BX,0EC00H
	JNC FILETOOBIG
	MOV [0FE00H+14],BX	;POKE IT INTO THE SIZE TO READ

	MOV DX,100H
	MOV [DMA],DX		;WE'LL READ IT INTO CPM LOCATION 100H
; READ A RECORD
READLOOP:
	MOV DX,[DMA]
	MOV AH,1AH	;SET DMA
	INT 21H

	MOV DX,0FE00H
	MOV AH,14H
	INT 21H		;READ PRG IN




; CREATE FIRST AND SECOND FCB IN PROGRAM AREA
;
; CLEAR FCB AREA
;
	MOV BX,5CH	;TARGET AREA IN DS
	MOV CX,24H	;COUNT 80H-5CH
CLRLOOP:
	MOV B,[BX],00H	;CLEAR A BYTE
	INC BX
	DEC CX
	JNZ CLRLOOP

;SET UP FIRST FILE NAME FOR EMULATOR
	MOV CX,16
	MOV BX,6CH
	MOV DI,5CH
MOVFILNAM:
	SEG CS
	MOV AL,[BX]
	MOV [DI],AL
	INC BX
	INC DI
	DEC CX
	JNZ MOVFILNAM

;NOW GEN THE SECOND FILENAME IN FCB

	MOV BX,81H
	MOV CL,[80H]
	MOV AL,[BX]
	CMP AL,20H
	JNZ SKIPFILE
	INC BX
	DEC CL

SKIPFILE:
	OR CL,CL
	JZ EMPTY
	MOV AL,[BX]
	CMP AL,20H
	JZ SKIPSPACE
	DEC CL
	INC BX
	JMP SKIPFILE

SKIPSPACE:
	INC BX
	DEC CL
	JZ EMPTY
	CMP B,[BX],20H
	JZ SKIPSPACE

;BX NOW POINTS TO BEG OF SECOND FILENAME,
; SO PARSE IT WITH MSDOS CALL 
;
FINDFILE:
	MOV DI,6CH
	MOV SI,DS
	MOV ES,SI
	MOV SI,BX
	MOV AL,00H
	MOV AH,29H
	INT 21H

	JMP FILEDONE

;GENERATE A BLANK FCB HERE IF NO NAME WAS GIVEN

EMPTY:
	MOV B,[6CH],00H

	MOV CL,11
	MOV BX,6DH
EMPTYSP:
	MOV B,[BX],20H
	INC BX
	DEC CL
	JNZ EMPTYSP
	JMP FILEDONE

;THE TRAILING NULLS WERE LOADED DURING THE FCB INIT



;THIS PRINTS A FILENAME FOR THE ERROR MSGS

FPRINT:
	PUSH BX
	PUSH DX
	PUSH CX
	MOV BX,DX
	MOV AL,[BX]
	INC BX
	CMP AL,00H
	JNZ FPRINTOK
	MOV AH,19H
	INT 21H
	INC AL
FPRINTOK:
	OR AL,40H
	CALL TYPEITOUT
	MOV AL,3AH
	CALL TYPEITOUT
	MOV CX,9
FPRINT1LP:
	MOV AL,[BX]
	INC BX
	DEC CX
	JZ FPRINT3
	CMP AL,' '
	JZ FPRINT2
	CALL TYPEITOUT
	JMP FPRINT1LP
FPRINT2:
	INC BX
	DEC CX
	JNZ FPRINT2
FPRINT3:
	DEC BX
	MOV AL,'.'
	CALL TYPEITOUT
	MOV CX,4
FPRINT3LP:
	MOV AL,[BX]
	INC BX
	DEC CX
	JZ FPRINT4
	CMP AL,' '
	JZ FPRINT4
	CALL TYPEITOUT
	JMP FPRINT3LP
FPRINT4:
	POP DX
	POP BX
	POP CX
	RET


;PRINT THE CHARACTER IN AL
TYPEITOUT:
	PUSH AX
	PUSH CX
	PUSH DX
	MOV DL,AL
	MOV AH,2
	INT 21H
	POP DX
	POP CX
	POP AX
	RET

;WE ARE ALL READY, SO LET'S GO OK?

FILEDONE:
	JMP PROGSTART


;********************************************************
; SET DMA TO CP/M DEFAULT

PROGSTART:
	MOV DX,80H
	MOV [DMA],DX
	MOV AH,1AH	;SET DMA
	INT 21H

	MOV BP,CPMSTACK	;INIT CPM STACK
	MOV [CPMSTACK],0FA00H	;EXIT TO COLD BOOT 


; TURN OFF TRACE
	MOV B,[BUGFLG],00H	;DOESN'T REALLY DO ANYTHING ANYMORE.

;*************************************************
;GET READY....

SETPC:
	PUSH DX		;PREP STACK FOR THE ENTRY
	PUSH AX
	PUSHF
	MOV DX,100H	;ENTER CODE AT 100H
	MOV BP,100H 	;INIT 8080 STACK TO 100H
	JMP EN8080	;GO PROCESS THE 8080 CODE

; EXTERNALS JUMPED TO SHOULD JUMP BACK TO DONE, WHEN FINISHED
; TO PERFORM THE NEXT 8080 INSTRUCTION.

DONE:
	PUSH DX		;WARNING: DS *MUST* BE PROPERLY SET 
	PUSH AX		;TO THE 8080 CODE SEGMENT.
	PUSHF
	SEG DS
	MOV DX,[BP]	;GET RETURN ADDRESS FROM 8080 STACK
	ADD BP,2	;POP OFF THE ADDRESS
EN8080:
	MOV AL,EMINT	;POKE THE 8080 RETURN ADDRESS INTO 
	MOV AH,25H	;THE EMULATION VECTOR
	INT 21H
	POPF
	POP AX
	POP DX
	DW BRKEM	;RETURN TO 8080 CODE
	DB EMINT
	JMP RBIOS	;GO HANDLE BIOS/BDOS REQUEST
;
;*******************************************************
;ERROR EXITS
;THIS IS REALLY ERROR STUFF LEFT OVER FROM THE SOFTWARE 8080 EMULATOR
;I'M NOT SURE WHY I LEFT IT IN OTHER THAN TO GIVE THOSE TO LIKE TO
;DIG INTO THE CODE SOMETHING TO SCRATCH THEIR HEAD ABOUT.

Z80MSG:
	DB ' <- Invalid Operation Code encountered ***',BELL,CR,LF,'$'


;THIS PRINTS THE DEBUG HEADER INFO SO WE KNOW WHAT ALL THAT JUNK MEANS
HEADER:
	PUSH AX
	PUSH CX
	PUSH DX
	PUSH DI
	PUSH SI
	PUSH BP
	PUSH BX
	PUSH DS
	MOV BX,CS
	MOV DS,BX
	MOV DX,HEADERMSG
	MOV AH,09H
	INT 21H
	POP DS
	POP BX
	POP BP
	POP SI
	POP DI
	POP DX
	POP CX
	POP AX
	RET

HEADERMSG:
	DB CR,LF,' PC   AF   BC   DE   HL   SP   (SP)  (SP-2)'
	DB ' OP-2 OP-1 OP OP+1 OP+2 OP+3$' 


;THIS PRINTS OUT THE 8080 REGISTERS AND STACK INFO ON ERRORS
DEBUG:
	PUSH AX
	PUSHF
	PUSH BX
	PUSH CX
	PUSH DX
	PUSH BP
	PUSH DI
	PUSH SI

	PUSH BP
	PUSH BX
	PUSH DX
	PUSH CX
	PUSH AX
	PUSH SI

	MOV DL,CR
	MOV AH,2
	INT 21H
	MOV DL,LF
	MOV AH,2
	INT 21H

	POP SI
	MOV BX,SI
	CALL PRTWORD
	POP AX
	MOV BH,AH
	MOV BL,AL
	CALL PRTWORD
	POP CX
	MOV BX,CX
	CALL PRTWORD
	POP DX
	MOV BX,DX
	CALL PRTWORD
	POP BX
	CALL PRTWORD
	POP BP
	MOV BX,BP
	CALL PRTWORD

	MOV DL,'('
	MOV AH,2
	INT 21H
	SEG DS
	MOV BX,[BP]
	MOV AL,BH
	CALL PRTBYTE
	MOV AL,BH
	CALL PRTBYTE
	MOV DL,')'
	MOV AH,2
	INT 21H
	MOV DL,' '
	MOV AH,2
	INT 21H

	MOV DL,'('
	MOV AH,2
	INT 21H
	SEG DS
	MOV BX,[BP+2]
	CALL PRTBYTE
	MOV AL,BL
	CALL PRTBYTE
	MOV DL,')'
	MOV AH,2
	INT 21H
	MOV DL,' '
	MOV AH,2
	INT 21H


	MOV DL,'('
	MOV AH,2
	INT 21H
	MOV AL,[SI-3]
	CALL PRTBYTE	;OPCODE MINUS TWO
	MOV DL,')'
	MOV AH,2
	INT 21H
	MOV DL,' '
	MOV AH,2
	INT 21H

	MOV DL,'('
	MOV AH,2
	INT 21H
	MOV AL,[SI-2]	;OP CODE MINUS ONE
	CALL PRTBYTE
	MOV DL,')'
	MOV AH,2
	INT 21H
	MOV DL,' '
	MOV AH,2
	INT 21H

	MOV AL,[SI-1]
	CALL PRTBYTE	;OP CODE
	MOV DL,' '
	MOV AH,2
	INT 21H


	MOV DL,'('
	MOV AH,2
	INT 21H
	MOV AL,[SI]
	CALL PRTBYTE	;OP CODE PLUS ONE
	MOV DL,')'
	MOV AH,2
	INT 21H
	MOV DL,' '
	MOV AH,2
	INT 21H

	MOV DL,'('
	MOV AH,2
	INT 21H
	MOV AL,[SI+1]
	CALL PRTBYTE	;OP CODE PLUS TWO
	MOV DL,')'
	MOV AH,2
	INT 21H
	MOV DL,' '
	MOV AH,2
	INT 21H

	MOV DL,'('
	MOV AH,2
	INT 21H
	MOV AL,[SI+2]
	CALL PRTBYTE	;OP CODE PLUS THREE
	MOV DL,')'
	MOV AH,2
	INT 21H
	MOV DL,' '
	MOV AH,2
	INT 21H

	POP SI
	POP DI
	POP BP
	POP DX
	POP CX
	POP BX
	POPF
	POP AX
	RET

PRTWORD:
	PUSH SI
	MOV AL,BH
	CALL PRTBYTE
	MOV AL,BL
	CALL PRTBYTE
	MOV AH,2
	MOV DL,' '
	INT 21H
	POP SI
	RET

PRTBYTE:
	PUSH AX
	PUSH DX
	PUSH BX
	MOV BL,AL
	ROR BL
	ROR BL
	ROR BL
	ROR BL
	AND BL,0FH
	CALL HEX2A
	MOV DL,BL
	PUSH AX
	MOV AH,2
	INT 21H
	POP AX
	MOV BL,AL
	AND BL,0FH
	CALL HEX2A
	MOV DL,BL
	MOV AH,2
	INT 21H
	POP BX
	POP DX
	POP AX
	RET

HEX2A:
	ADD BL,30H
	CMP BL,3AH
	JC HEXN
	ADD BL,07H
HEXN:	RET


	DB CR,LF,'This program was written by: Mike Day',CR,LF,'$'


;PROGRAM TEMPORARY FCB TO READ COM FILE

FCB:
	DB 0,0,0,0,0,0,0
	DB 0,0,0,0,0,0,0
	DB 0,0,0,0,0,0,0
	DB 0,0,0,0,0,0,0
	DB 0,0,0,0,0,0,0
	DB 0,0,0,0



; MAP OUT REAL CPM SYSTEM


BOOT:	DB 0C3H,03H,0FAH

IOBYTE:	DB 094H

USRBYT:	DB 00H

BDOS:	DB 0C3H,06H,0ECH
	DS 54H
	DB 00H,20H,20H,20H
	DS 0A0H
TBA:
	DB 76H



; DECODE BDOS CALLS

REALBDOS:			;0EC06H
	DW 7F3EH	;MVI A,7FH
	DW RETEM	;RBDOS
	DB 68H
	DB 0C9H


; CPM JUMP TABLE
; (FOR THOSE WHO NEED TO SEE A JUMP TABLE HERE) 
;
CPMJMP:
	DB 0C3H
	DW 0FB00H	;CBOOT

	DB 0C3H
	DW 0FB06H	;WBOOT

	DB 0C3H
	DW 0FB0CH	;CONST

	DB 0C3H
	DW 0FB12H	;CONIN

	DB 0C3H
	DW 0FB18H	;CONOUT

	DB 0C3H
	DW 0FB1EH	;LIST

	DB 0C3H
	DW 0FB24H	;PUNCH

	DB 0C3H
	DW 0FB2AH	;READER

	DB 0C3H
	DW 0FB30H	;HOME

	DB 0C3H
	DW 0FB36H	;SELDSK

	DB 0C3H
	DW 0FB3CH	;SETTRK

	DB 0C3H
	DW 0FB42H	;SETSEC

	DB 0C3H
	DW 0FB48H	;SETDMA

	DB 0C3H
	DW 0FB4EH	;READ

	DB 0C3H
	DW 0FB54H	;WRITE

	DB 0C3H
	DW 0FB5AH	;LISTST

	DB 0C3H
	DW 0FB60H	;SECTRAN


;NOW CONVERT TO OUR OWN CONTROL VIA MACRO COMMAND

CBOOT:	DW 013EH 	;MVI A,01H
	DW RETEM		;0FB00H
	DB 69H
	DB 0C9H

WBOOT:	DW 023EH
	DW RETEM		;0FB03H
	DB 69H
	DB 0C9H

CONST:	DW 033EH
	DW RETEM		;0FB06H
	DB 69H
	DB 0C9H

CONIN:	DW 043EH
	DW RETEM		;0FB09H
	DB 69H
	DB 0C9H

CONOUT:	DW 053EH
	DW RETEM		;0FB0CH
	DB 69H
	DB 0C9H

LIST:	DW 063EH
	DW RETEM		;0FB0FH
	DB 69H
	DB 0C9H

PUNCH:	DW 073EH
	DW RETEM		;0FB12H
	DB 69H
	DB 0C9H

READER:	DW 083EH
	DW RETEM		;0FB15H
	DB 69H
	DB 0C9H

HOME:	DW 093EH
	DW RETEM		;0FB18H
	DB 69H
	DB 0C9H

SELDSK:	DW 0A3EH
	DW RETEM		;0FB1BH
	DB 69H
	DB 0C9H

SETTRK:	DW 0B3EH
	DW RETEM		;0FB1EH
	DB 69H
	DB 0C9H

SETSEC:	DW 0C3EH
	DW RETEM		;0FB21H
	DB 69H
	DB 0C9H

SETDMA:	DW 0D3EH
	DW RETEM		;0FB24H
	DB 69H
	DB 0C9H

READ:	DW 0E3EH
	DW RETEM		;0FB27H
	DB 69H
	DB 0C9H

WRITE:	DW 0F3EH
	DW RETEM		;0FB2AH
	DB 69H
	DB 0C9H

LISTST:	DW 103EH
	DW RETEM		;0FB5AH
	DB 69H
	DB 0C9H

SECTRAN: DW 113EH
 	DW RETEM		;0FB60H
	DB 69H
	DB 0C9H

	DB 'With us to the bitter end huh? For 40 bucks you could easily',CR,LF
	DB 'find out what that junk above was all about.',CR,LF
	DB 'Just so you will have it handy, the address is',CR,LF
	DB 'Mike Day, C/O Day Research, P.O. Box 22902, Milwaukie, OR 97222',CR,LF
	DB 'Remember, for only 40 bucks you get the manual',CR,LF 
	DB '*and* the commented source code. How can you pass up',CR,LF
	DB 'a deal like that?',CR,LF 
	DB 'Make your ',24h,'40 check payable to Michael Day',CR,LF
	DB CR,LF,'$'

;*****************************************
; END OF BIOS SECTION
;***************************************
;
;
; STACK AREA HERE
;
	DS 400H
STACK	EQU $
;
;
;
;
;  8080 REGISTERS MAPPED INTO MEMORY
;
	DS 0FC00H-$
;
;
I2BC:	DS 2	;left over from software emulator
I2DE:	DS 2
I2HL:	DS 2
I2PSW:	DS 2

IIX:	DS 2	;left over from z80 emulator
IIY:	DS 2
IIR:
IR:	DS 1
II:	DS 1
;
	DS 2	;EXTRA
DMA	DS 2	;DMA ADDRESS
;
CPMEND:
	DS 2

SEG8080	DS 2 	;data segment address storage for 8080 code segment


LASTXY	DS 2	;ADM3A EMULATION VARS
CURFLAG	DS 2
EQFLAG	DS 2
ROWFLAG DS 2
ROWCNT	DS 2


BUGFLG:
	DS 1
	DS 1	;SAFETY
;

BYTEIN:
	DS 1

BYTERDYFLG:
	DS 1
	DS 1
COMBRDYFLG:
	DS 1
	DS 1

COMBYTEIN:
	DS 1
	DS 1
GETIN:
	DS 1
	DS 1
STRDY:
	DS 1
	DS 1


PRTFLAG:
	DS 1
	DS 1
SPCOUNT:
	DS 1
	DS 1

DRVLOG:
	DS 2	;DRV SEL FLAG


CONFLAG:
	DS 1
	DS 1

BEGBUFF:
	DS 1
	DS 1

DELCOUNT:
	DS 1
	DS 1

DISKREG:
	DS 1
	DS 1

DMAFLAG:
	DS 1
	DS 1

OLDBUGVAL:
	DS 1
	DS 1


;
TEMPFCB:
	DS 48
	DS 32
DPARM:
	DS 16
ALLOC:
	DS 256

	DS 32	;CPM STACK
CPMSTACK:
	DS 4	;BEG OF CPM STACK


	END 100H
